@@ -1,4 +1,13 @@
 ﻿# Collective Mind automation
+#
+# Author(s): Grigori Fursin
+# Contributor(s):
+#
+# Copyright: https://github.com/mlcommons/ck/blob/master/COPYRIGHT.txt
+# License: https://github.com/mlcommons/ck/blob/master/LICENSE.md
+# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
+# White paper: https://arxiv.org/abs/2406.16791
+# Project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
 
 import os
 
@@ -120,9 +129,9 @@ class Automation:
 
         print ('')
         print ('GitHub for CM developments:        https://github.com/mlcommons/ck/tree/master/cm')
-        print ('GitHub for CM automation scripts:  https://github.com/mlcommons/ck/tree/master/cm-mlops')
+        print ('GitHub for CM automation scripts:  https://github.com/mlcommons/cm4mlops')
         print ('Reporting issues and ideas:        https://github.com/mlcommons/ck/issues')
-        print ('Joining the open MLPerf workgroup: https://cKnowledge.org/mlcommons-taskforce')
+        print ('MLCommons taskforce developing CM: https://github.com/mlcommons/ck/blob/master/docs/taskforce.md')
 
         return {'return':0}
 
@@ -260,11 +269,9 @@ class Automation:
                             print (artifact_object.path)
 
                     return {'return':0, 'list':lst}
-        
-        
-        
 
-        
+
+
         # Iterate over pruned repositories
         for repo in pruned_repos:
             path_repo = repo.path_with_prefix
@@ -293,7 +300,7 @@ class Automation:
                         path_artifact = os.path.join(path_artifacts, artifact)
 
                         if os.path.isdir(path_artifact):
-                            
+
                             # Check if has CM meta to make sure that it's a CM object
                             path_artifact_meta = os.path.join(path_artifact, self.cmind.cfg['file_cmeta'])
 
@@ -396,6 +403,7 @@ class Automation:
             meta (dict): meta description of this artifact
 
             (tags) (string or list): tags to be added to meta
+            (new_tags) (string or list): new tags to be added to meta (the same as tags)
             (yaml) (bool): if True, record YAML instead of JSON
 
         Returns: 
@@ -483,9 +491,20 @@ class Automation:
         if meta.get('automation_uid','')=='': meta['automation_uid']=automation_name[1]
 
         existing_tags = meta.get('tags',[])
-        if len(tags)>0: 
-            existing_tags += tags
-        meta['tags'] = utils.filter_tags(existing_tags)
+
+        new_tags_list = []
+        if i.get('new_tags','')!='':
+            new_tags_list = i['new_tags'].split(',')
+
+        if len(tags)>0:
+            new_tags_list += tags    
+
+        if len(new_tags_list)>0: 
+            for xtag in new_tags_list:
+                if xtag!='' and xtag not in existing_tags:
+                    existing_tags.append(xtag)
+
+            meta['tags'] = existing_tags
 
         # Record meta
         if i.get('yaml', False):
@@ -566,7 +585,10 @@ class Automation:
             path_to_artifact = artifact.path
 
             if console:
-                print ('Deleting CM artifact in {} ...'.format(path_to_artifact))
+                tags = artifact.meta.get('tags',[])
+                x = '' if len(tags)=='' else ' with tags "{}"'.format(','.join(tags))
+                
+                print ('Deleting CM artifact in {}{} ...'.format(path_to_artifact, x))
 
                 if not force:
                     ask = input('  Are you sure you want to delete this artifact (y/N): ')
@@ -584,7 +606,8 @@ class Automation:
 
             if os.name == 'nt':
                 # To be able to remove .git files
-                shutil.rmtree(path_to_artifact, ignore_errors = False, onerror = delete_helper)
+                shutil.rmtree(path_to_artifact, onerror = utils.rm_read_only)
+#                shutil.rmtree(path_to_artifact, ignore_errors = False, onerror = delete_helper)
             else:
                 shutil.rmtree(path_to_artifact)
 
@@ -658,8 +681,12 @@ class Automation:
 
         # Output if console
         if console:
-            import json
-            print (json.dumps(meta, indent=2, sort_keys=True))
+            if str(i.get('yaml','')).lower() in ['true','yes']:
+                import yaml
+                print (yaml.dump(meta))
+            else:
+                import json
+                print (json.dumps(meta, indent=2, sort_keys=True))
 
         return {'return':0, 'path':path, 'meta':meta, 'original_meta':original_meta, 'artifact':artifact}
 
@@ -970,6 +997,8 @@ class Automation:
                                       [ (artifact alias, artifact UID) ] or
                                       [ (artifact alias, artifact UID), (artifact repo alias, artifact repo UID) ]
 
+            new_tags (string): add new tags (separated by comma) to new artifacts
+        
         Returns: 
             (CM return dict):
 
@@ -1030,6 +1059,11 @@ class Automation:
             target_repo_path = os.path.abspath(target_artifact_repo_obj.path_with_prefix)
 
         # Updating artifacts
+        new_tags_list = []
+        if i.get('new_tags','')!='':
+            new_tags_list = i['new_tags'].split(',')
+
+        
         for artifact in lst:
 
             artifact_path = os.path.abspath(artifact.path)
@@ -1059,6 +1093,12 @@ class Automation:
                 artifact_meta['alias']=target_artifact_obj_alias
 
             artifact_meta['uid']=target_artifact_obj_uid
+            if len(new_tags_list)>0:
+                xtags = artifact_meta.get('tags',[])
+                for xtag in new_tags_list:
+                    if xtag!='' and xtag not in xtags:
+                        xtags.append(xtag)
+                artifact_meta['tags'] = xtags
 
             artifact.path = new_artifact_path
 
@@ -1162,6 +1202,8 @@ class Automation:
 
         cid1 = ''
         
+    
+
         for artifact in lst:
             if console and len(lst)>1:
                 print ('****************************************************')
@@ -1196,9 +1238,26 @@ class Automation:
                 print ('CID = ' + cid)
                 print ('CID1 = ' + cid1)
                 print ('CID2 = ' + full_cid)
+                print ('')
                 print ('Path = ' + path)
 
-        # Attempt to copy to clipboard the last CID
+                p_dirs=[]
+                p_files=[]
+                for p in os.listdir(path):
+                    p_dirs.append(p) if os.path.isdir(os.path.join(path,p)) else p_files.append(p)
+                        
+                if len(p_dirs)>0 or len(p_files)>0:
+                    for x in [('Directories',p_dirs), 
+                              ('Files',p_files)]:
+                        x0 = x[0]
+                        x1 = x[1]
+                        if len(x1)>0:
+                            print ('')
+                            print ('  '+x0+':')
+                            for p in sorted(x1):
+                                print ('    ' + p)
+
+       # Attempt to copy to clipboard the last CID
         if cid1 !='':
             clipboard = full_cid
             if i.get('uid', False): clipboard = cid1
@@ -1210,15 +1269,15 @@ class Automation:
 
         return {'return':0, 'list': lst}
 
-############################################################
-def delete_helper(func, path, ret):
-    import stat, errno
-
-    if ret[1].errno != errno.EACCES:
-        raise
-    else:
-        clean_attr = stat.S_IRWXG | stat.S_IRWXO | stat.S_IRWXU
-        os.chmod(path, clean_attr)
-        func(path)
-
-    return
+#############################################################
+#def delete_helper(func, path, ret):
+#    import stat, errno
+#
+#    if ret[1].errno != errno.EACCES:
+#        raise
+#    else:
+#        clean_attr = stat.S_IRWXG | stat.S_IRWXO | stat.S_IRWXU
+#        os.chmod(path, clean_attr)
+#        func(path)
+#
+#    return
