@@ -1,4 +1,13 @@
 ﻿# Collective Mind configuration
+#
+# Author(s): Grigori Fursin
+# Contributor(s):
+#
+# Copyright: https://github.com/mlcommons/ck/blob/master/COPYRIGHT.txt
+# License: https://github.com/mlcommons/ck/blob/master/LICENSE.md
+# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
+# White paper: https://arxiv.org/abs/2406.16791
+# Project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
 
 import os
 
@@ -35,7 +44,9 @@ class Config(object):
                      "flag_help2": "help",
 
                      "error_prefix": "CM error:",
-                     "info_cli": "cm {action} {automation} {artifact(s)} {--flags} @input.yaml @input.json",
+                     "error_prefix2": "CMX detected a problem",
+                     "info_cli": "cm {action} {automation} {artifact(s)} {flags} @input.yaml @input.json",
+                     "info_clix": "cmx {action} {automation} {artifact(s)} {CMX control flags (-)} {CMX automation flags (--)}",
 
                      "default_home_dir": "CM",
 
@@ -48,6 +59,7 @@ class Config(object):
                      "file_meta_repo": "cmr",
 
                      "common_automation_module_name": "module",
+                     "common_automation_module_namex": "modulex",
 
                      "action_substitutions": {
                        "ls":"search",
@@ -60,7 +72,7 @@ class Config(object):
                        "cp":"copy"
                      },
 
-                     "new_repo_requirements": "cmind >= 0.7.5\n",
+                     "new_repo_requirements": "cmind >= 3.1.0\n",
 
                      "cmind_automation":"automation",
 
