@@ -1,4 +1,13 @@
 ﻿# Collective Mind repository
+#
+# Author(s): Grigori Fursin
+# Contributor(s):
+#
+# Copyright: https://github.com/mlcommons/ck/blob/master/COPYRIGHT.txt
+# License: https://github.com/mlcommons/ck/blob/master/LICENSE.md
+# History: https://github.com/mlcommons/ck/blob/master/HISTORY.CM.md
+# White paper: https://arxiv.org/abs/2406.16791
+# Project contributors: https://github.com/mlcommons/ck/blob/master/CONTRIBUTING.md
 
 import os
 
@@ -37,9 +46,10 @@ class Repo:
 
         # Repo meta
         self.meta = {}
+        self.original_meta = {} # for compatibility with CM/CMX universal search
 
     ############################################################
-    def load(self):
+    def load(self, cmx = False):
         """
         Load CM repository
 
@@ -63,13 +73,22 @@ class Repo:
 
         r = utils.load_yaml_and_json(file_name_without_ext = full_path)
         if r['return'] >0: 
-           r['error']='CM repository is broken ({})'.format(r['error'])
+           r['error'] = 'CM repository is broken ({})'.format(r['error'])
+           r['return'] = 16
            return r
 
         self.meta = r['meta']
+        self.original_meta = self.meta
+
+        if cmx and self.meta.get('prefix_cmx', '') != '':
+            self.path_prefix = self.meta.get('prefix_cmx','')
+        else:
+            self.path_prefix = self.meta.get('prefix','')
 
-        self.path_prefix = self.meta.get('prefix','')
         if self.path_prefix is not None and self.path_prefix.strip() !='':
             self.path_with_prefix = os.path.join(self.path, self.path_prefix)
 
+        if os.path.isdir(self.path) and not os.path.isdir(self.path_with_prefix):
+            os.makedirs(self.path_with_prefix)
+
         return {'return':0}
